package handlers

import (
	"context"
	"errors"
	"fmt"
	"net/http"

	"github.com/Masterminds/squirrel"
	"github.com/go-openapi/runtime/middleware"
	"github.com/jmoiron/sqlx"
	"github.com/rs/zerolog"
	"orus.io/orus-io/go-orusapi/database"
	"orus.io/orus-io/rednerd/models"
	"orus.io/orus-io/rednerd/rendering"
	op "orus.io/orus-io/rednerd/restapi/operations/rendering"
	"orus.io/orus-io/rednerd/utils"
)

// NewRenderHandler creates a RenderHandler
func NewRenderHandler(
	db *sqlx.DB,
	registry *rendering.EngineRegistry,
	templateRegistry *rendering.TemplateEngineRegistry,
) *RenderHandler {
	return &RenderHandler{db, registry, templateRegistry}
}

// RenderHandler handles POST /api/v1/render
type RenderHandler struct {
	db               *sqlx.DB
	registry         *rendering.EngineRegistry
	templateRegistry *rendering.TemplateEngineRegistry
}

func (h *RenderHandler) handle(
	ctx context.Context,
	template *models.Template,
	data models.Dataset,
	document *models.Document,
	metadata models.Metadata,
	toType string,
) ([]*models.Document, error) {
	if template == nil && document == nil {
		return nil, utils.HTTPBadRequest(
			errors.New("no template nor document. Must have one"))
	}

	var (
		log    = zerolog.Ctx(ctx)
		result []*models.Document
	)

	if template != nil {
		if template.Account != "" && template.Name != "" {
			templatedb := models.TemplateDB{}
			sql := database.NewSQLHelper(ctx, h.db, *log)
			if err := sql.GetWhere(
				&templatedb,
				squirrel.Eq{
					models.TemplateDBAccountColumn: template.Account,
					models.TemplateDBNameColumn:    template.Name,
				},
			); err != nil {
				return nil, err
			}
			templatedb.FromDB()
			template = &templatedb.Template
		}
		pipeline, err := h.templateRegistry.NewPipeline(
			ctx, h.registry, template, metadata, toType,
		)
		if err != nil {
			if err == rendering.ErrUnknownTemplateLanguage {
				return nil, utils.HTTPBadRequest(
					fmt.Errorf("unknown template language: %s", template.Language))
			}
			return nil, utils.HTTPBadRequest(
				fmt.Errorf("error loading rendering pipeline: %s", err))
		}
		for _, record := range data {
			var doc models.Document
			w := rendering.NewDocumentWriter(&doc)
			if err := pipeline.Render(ctx, w, record); err != nil {
				return nil, err
			}
			result = append(result, &doc)
		}
	}

	if document != nil {
		pipeline, err := h.registry.GetPipeline(document.Type, toType)
		if err != nil {
			return nil, utils.HTTPBadRequest(err)
		}

		document.UpdateMetadata(metadata)

		if len(pipeline) != 0 {
			var output models.Document

			reader := rendering.NewDocumentReader(document)
			writer := rendering.NewDocumentWriter(&output)

			if err := pipeline.Render(ctx, writer, reader); err != nil {
				return nil, err
			}
			result = append(result, &output)
		} else {
			result = append(result, document)
		}
	}

	return result, nil
}

// Handle the incoming request
func (h *RenderHandler) Handle(params op.RenderParams, principals *models.Principal) middleware.Responder {
	accept := params.Request.Accept
	// TODO if accept is empty, get a 'accept' from the http header

	// TODO checks if the user has render access to the template

	docs, err := h.handle(
		params.HTTPRequest.Context(),
		params.Request.Template,
		params.Request.Data,
		params.Request.Document,
		params.Request.Metadata,
		string(accept),
	)

	if err != nil {
		if he, isHTTPError := err.(utils.HTTPError); isHTTPError {
			r := op.NewRenderDefault(he.StatusCode)
			if he.Err != nil {
				r.WithPayload(models.FromError(he.Err))
			}
			return r
		}
		return op.
			NewRenderDefault(http.StatusInternalServerError).
			WithPayload(models.FromError(err))
	}

	return op.NewRenderOK().WithPayload(docs)
}
